/* ScaFES
 * Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 * For details, see the files COPYING and LICENSE in the base directory
 * of the package.
 */

/**
 *  @file ScaFES_Parameters.hpp
 *  @brief Contains the class Parameters.
 */

#ifndef SCAFES_PARAMETERS_HPP_
#define SCAFES_PARAMETERS_HPP_

#include "ScaFES_Config.hpp"
#include "ScaFES_Communicator.hpp"

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <vector>
#include <string>

#include <boost/serialization/serialization.hpp>

// #ifdef SCAFES_HAVE_BOOST_SERIALIZATION
// namespace boost
// {
// namespace serialization
// {
// class access;
// }
// }
// #include <boost/serialization/vector.hpp>
// #include <boost/serialization/version.hpp>
// #endif


#ifdef SCAFES_HAVE_BOOST_PROGRAM_OPTIONS
#include <boost/program_options.hpp>
#endif

#ifdef SCAFES_HAVE_BOOST
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#endif

#ifdef SCAFES_HAVE_BOOST_REGEX
#include <boost/regex.hpp>
#endif

#include "ScaFES_Ntuple.hpp"

namespace ScaFES
{

#ifdef SCAFES_HAVE_BOOST_PROGRAM_OPTIONS
namespace po = boost::program_options;
#endif

/*******************************************************************************
 * GLOBAL VARIABLES.
 ******************************************************************************/
/** Maximal indentation depth level of console output. */
int gOutputLevelMax = 2;

/** Current indentation depth level of console output. */
int gOutputLevelCurr = gOutputLevelMax;

/*******************************************************************************
 ******************************************************************************/
/**
 * \class Parameters
 * @brief The class \c Parameters contains all parameters related to ScaFES.
 *
 * Parameters can be given via the command line.
 * If a command line parameter is invalid, a help message will be displayed.
 *
 * The MPI environment lives as long as an instance of this class is living.
 * In particular, if the destructor of this class is called,
 * the MPI environment will be finalized. Thus, if the descructor is called
 * too early, the instance of the calls goes out of scope. So, please be careful
 * with calls to the destructor of this class.
 *
 * \remarks The master MPI process reads in the command line arguments
 * and distributed the parameters to all other MPI processes.
 */
class Parameters
{
public:
    /*----------------------------------------------------------------------
    | FRIEND CLASSES.
    ----------------------------------------------------------------------*/
    friend class boost::serialization::access;

    /*----------------------------------------------------------------------
    | LIFE CYCLE METHODS.
    ----------------------------------------------------------------------*/
    /** Creates own default constructor. */
    Parameters();

    /** Creates own constructor from command line parameters.
     * Parses the command line of a program call and provides a help
     * message.
     */
    Parameters(int argc, char* argv[]);

    /** Creates own copy constructor. */
    Parameters(Parameters const&);

    /** Creates own destructor. */
    ~Parameters();

    /** Creates own assignment operator. */
    Parameters& operator=(Parameters rhs);

    /*----------------------------------------------------------------------
    | GETTER METHODS.
     ----------------------------------------------------------------------*/
    /** Returns the space dimension. */
    int const& dim() const;

    /** Returns the number of grid nodes in each dimension. */
    std::vector<int> const& nNodes() const;

    /** Returns the coordinates of the first grid point. */
    std::vector<double> const& coordNodeFirst() const;

    /** Returns the coordinates of the last grid point. */
    std::vector<double> const& coordNodeLast() const;

    /** Returns the number of partitions in each direction. */
    std::vector<int> const& nPartitions() const;

    /** Returns the directions in which the grid should be partitioned. */
    std::vector<bool> const& divideGrid() const;

    /** Returns the decision if the kind field should be read in. */
    bool const& readKindFile() const;

    /** Returns if the kind field should be written to file. */
    bool const& writeKindFile() const;

    /** Returns the name of the kind file. */
    std::string const& kindFile() const;

    /** Returns the name of the domain decomposition approach. */
    std::string const& typeDomainDecomp() const;

    /** Returns the name of the partition file for writing. */
    std::string const& nameWritePartitionFile() const;

    /** Returns the decision if a partition file should be written. */
    bool const& writePartitionFile() const;

    /** Returns the name of the partition file for reading. */
    std::string const& nameReadPartitionFile() const;

    /** Returns the decision if a partition file should be written. */
    bool const& readPartitionFile() const;

    /** Returns the start of the time interval. */
    double const& timeIntervalStart() const;

    /** Returns the end of the time interval. */
    double const& timeIntervalEnd() const;

    /** Returns time step width. */
    double const& tau() const;

    /** Returns the number of time steps. */
    int const& nTimesteps() const;

    /** Returns the number of snapshots. */
    int const& nSnapshots() const;

    /** Returns the number of layers at the border. */
    int const& nLayersAtBorder() const;

    /** Returns the decision if an output file should be written. */
    bool const& writeDataFile() const;

    /** Returns the name of the output file. */
    std::string const& nameDataFile() const;

    /** Returns the name of the config file. */
    std::string const& nameConfigFile() const;

    /** Returns if ADOL-C is enabled. */
    bool const& enabledAdolc() const;

    /** Returns if asynchronous MPI communication is enabled. */
    bool const& asynchronMode() const;

    /** Returns if gradients should be computed. */
    bool const& computeGradients() const;

    /** Returns if gradients should be checked. */
    bool const& checkGradients() const;

    /** Returns the MPI rank for output. */
    int rankOutput() const;

    /** Returns the MPI communicator. */
    ScaFES::Communicator const& myWorld() const;

    /** Returns the MPI rank. */
    int rank() const;

    /** Returns the maximum level of console output. */
    int const& indentDepth() const;

    /*----------------------------------------------------------------------
    | SETTER METHODS.
     ----------------------------------------------------------------------*/
    /** Increase output level by one. */
    void increaseLevel();

    /** Decrease output level by one. */
    void decreaseLevel();

    /*----------------------------------------------------------------------
    | COMPARISON METHODS.
    ----------------------------------------------------------------------*/
    /** Tests if two grids are equal. */
    bool operator==(Parameters const& rhs) const;

    /** Tests if two grids are not equal. */
    bool operator!=(Parameters const& rhs) const;

    /*----------------------------------------------------------------------
    | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
    ----------------------------------------------------------------------*/
    /** Swaps to elements of this class. */
    friend void swap(Parameters& first, Parameters& second);

    /** Overloads the output operator \c operator<<. */
    friend std::ostream& operator<<(std::ostream& output, Parameters const& pp);

private:
    /*----------------------------------------------------------------------
    | WORK METHODS.
    ----------------------------------------------------------------------*/
    /** Checks a parameter given by its name. */
    template <typename TT>
    void checkParameter(po::variables_map const& vm,
                        std::string const& nameParam, std::vector<TT>& param,
                        bool isNecessary);

    /** Serializes class. */
    template <class Archive>
    void serialize(Archive& ar, const unsigned int version);

    /*----------------------------------------------------------------------
    | MEMBER VARIABLES.
    ----------------------------------------------------------------------*/
    /** Space dimension. */
    int mDim;

    /** Number of grid nodes in each dimension. */
    std::vector<int> mNnodes;

    /** Coordinate of first grid point (left lower point of grid). */
    std::vector<double> mCoordNodeFirst;

    /** Coordinate of last grid point (right upper point of grid). */
    std::vector<double> mCoordNodeLast;

    /** Number of partitions in each direction. */
    std::vector<int> mNpartitions;

    /** Divide grid into partitions: Which directions are enabled? */
    std::vector<bool> mDivideGrid;

    /** Read kind field from a given kind file? */
    bool mReadKindFile;

    /** Write kind field? */
    bool mWriteKindFile;

    /** Name of the kind file. */
    std::string mKindFile;

    /** Type of domain decomposition. */
    std::string mTypeDomainDecomp;

    /** Write partition table to a partition file? */
    bool mWritePartitionFile;

    /** Name of partition file which should be written. */
    std::string mNameWritePartitionFile;

    /** Read partition table from a partition file? */
    bool mReadPartitionFile;

    /** Name of the partition file which should be read in. */
    std::string mNameReadPartitionFile;

    /** Start of time interval [tstart;tend]. */
    double mTimeIntervalStart;

    /** End of time interval [tstart;t_end]. */
    double mTimeIntervalEnd;

    /** Time step width. */
    double mTau;

    /** Number of timesteps. */
    int mNtimesteps;

    /** Number of snapshots: Write data each m timesteps. */
    int mNsnapshots;

    /** Number of layers at the border. */
    int mNlayersAtBorder;

    /** Write output file? */
    bool mWriteDataFile;

    /** Name of the output file. */
    std::string mNameOfDataFile;

    /** Name of the configuration file. */
    std::string mNameOfConfigFile;

    /** Is ADOL-C enabled? */
    bool mEnabledAdolc;

    /** Is asynchronous MPI communication enabled? */
    bool mAsynchronMode;

    /** Should gradients be computed? */
    bool mComputeGradients;

    /** Should gradients be checked? */
    bool mCheckGradients;

    /** MPI rank for output. */
    int mRankOutput;

    /** MPI communictor. */
    ScaFES::Communicator mMyWorld;
}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
inline Parameters::Parameters()
: mDim(3)
, mNnodes(3)
, mCoordNodeFirst(3)
, mCoordNodeLast(3)
, mNpartitions(3)
, mDivideGrid(3)
, mReadKindFile(false)
, mWriteKindFile(false)
, mKindFile("EMPTY")
, mTypeDomainDecomp("RCB")
, mWritePartitionFile(false)
, mNameWritePartitionFile("EMPTY")
, mReadPartitionFile(false)
, mNameReadPartitionFile("EMPTY")
, mTimeIntervalStart(0.0)
, mTimeIntervalEnd(1.0)
, mTau(1.0)
, mNtimesteps(0)
, mNsnapshots(1)
, mNlayersAtBorder(1)
, mWriteDataFile(false)
, mNameOfDataFile("EMPTYDATAFILE")
, mNameOfConfigFile("EMPTYCONFIGFILE")
, mEnabledAdolc(false)
, mAsynchronMode(true)
, mComputeGradients(false)
, mCheckGradients(false)
, mRankOutput(0)
, mMyWorld()
{
    for (int ii = 0; ii < dim(); ++ii)
    {
        mNnodes.at(ii) = 8;
        mCoordNodeFirst.at(ii) = 0.0;
        mCoordNodeLast.at(ii) = 1.0;
        mDivideGrid.at(ii) = false;
    }
}
/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(int argc, char* argv[])
: mDim(3)
, mNnodes()
, mCoordNodeFirst()
, mCoordNodeLast()
, mNpartitions(3)
, mDivideGrid()
, mReadKindFile(false)
, mWriteKindFile(false)
, mKindFile("EMPTY")
, mTypeDomainDecomp("RCB")
, mWritePartitionFile(false)
, mNameWritePartitionFile("EMPTY")
, mReadPartitionFile(false)
, mNameReadPartitionFile("EMPTY")
, mTimeIntervalStart(0.0)
, mTimeIntervalEnd(1.0)
, mTau(0.0)
, mNtimesteps(0)
, mNsnapshots(1)
, mNlayersAtBorder(1)
, mWriteDataFile(false)
, mNameOfDataFile("EMPTYDATAFILE")
, mNameOfConfigFile("EMPTYCONFIGFILE")
, mEnabledAdolc(false)
, mAsynchronMode(true)
, mComputeGradients(false)
, mCheckGradients(false)
, mRankOutput(0)
, mMyWorld(argc, argv)
{
    if (0 == this->myWorld().rank())
    {

        po::variables_map vm;

        /*----------------------------------------------------------------------
        | Describe the options of the command line.
        ----------------------------------------------------------------------*/
        po::options_description desc("Allowed ScaFES options");
        desc.add_options()("help", "Displays this help message.")(
            "dim", po::value<int>(), "Sets the space dimension.")(
            "nPartitions", po::value<std::string>(),
            "Sets the number of partitions in each dimension (Np1x..xNpd).")(
            "typeDomainDecomposition", po::value<std::string>(),
            "Sets the type of domain decomposition: RCB (RCB algorithm), UNI "
            "(Uniform decomposition in each direction).")(
            "nNodes", po::value<std::string>(),
            "Sets the number of grid nodes in each dimension (N1x..xNd).")(
            "coordNodeFirst", po::value<std::string>(),
            "Sets the coordinates of the first grid node (c1x..xcd).")(
            "coordNodeLast", po::value<std::string>(),
            "Sets the coordinates of the last grid node (c1x..xcd).")(
            "divideGrid", po::value<std::string>(),
            "Divides the grid in which dimension? (g1x..xgd).")(
            "readKindFile", po::value<std::string>(),
            "Reads the kind field from a given kind file.")(
            "writeKindFile", po::value<int>(),
            "Writes the kind field to a NetCDF data file.")(
            "readPartitionfile", po::value<std::string>(),
            "Reads the partition table from a given partition file.")(
            "writePartitionfile", po::value<std::string>(),
            "Writes the partition table to a given partition file.")(
            "starttime", po::value<double>(), "Sets the start time.")(
            "endtime", po::value<double>(), "Sets the end time.")(
            "tau", po::value<double>(), "Sets the time step width.")(
            "nTimesteps", po::value<int>(),
            "Sets the number of time steps (=#{time intervals} (default=0).")(
            "nSnapshots", po::value<int>(), "Sets how many times the data "
                                            "fields are written to output "
                                            "(default=1).")(
            "nLayersAtBorder", po::value<int>(),
            "Sets of how many layers the border should consist (default=1).")(
            "outputfile", po::value<std::string>(),
            "Sets the name of the output data file.")(
            "configfile", po::value<std::string>(),
            "Sets the name of the config file.")(
            "enabledAdolc", po::value<int>(),
            "Sets if ADOL-C should be enabled (default=0)")(
            "asynchronMode", po::value<int>(), "Sets if asynchronous MPI "
                                               "communication should be used "
                                               "(default=1).")(
            "computeGradients", po::value<int>(),
            "Sets if gradients should be computed (default=0).")(
            "checkGradients", po::value<int>(), "Sets if gradients of forward "
                                                "computation should be checked "
                                                "(default=0).")(
            "rankOutput", po::value<int>(),
            "Sets the rank of the MPI process whose output should be displayed "
            "(default=0).")("outputlevel", po::value<int>(),
                            "Controls the amount of console output.");

        /*----------------------------------------------------------------------
        | Try to parse the command line.
        ----------------------------------------------------------------------*/
        try
        {
            po::parsed_options parsed =
                po::command_line_parser(argc, argv).options(desc).allow_unregistered().run();
            po::store(parsed, vm);
            po::notify(vm);
        }
        catch (std::exception& e)
        {
            throw std::runtime_error("Wrong command line options.");
        }

        /*----------------------------------------------------------------------
        | Actions of command line options.
        ----------------------------------------------------------------------*/
        if (vm.count("help"))
        {
            std::cout << desc << "\n";
        }
        else {

        /*----------------------------------------------------------------------
        | NECESSARY PARAMETERS.
        ----------------------------------------------------------------------*/
        if (vm.count("dim"))
        {
            this->mDim = vm["dim"].as<int>();

            if (0 >= this->dim())
            {
                std::cerr << "\nERROR: Space dimension is less or equal"
                             " to zero. Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nERROR: Use --dim=<integral value>.\n\n";
            this->mMyWorld.abort(-1);
        }

        /*--------------------------------------------------------------------*/
        this->mNnodes.resize(this->dim());
        this->checkParameter(vm, "nNodes", mNnodes, true);

        /*--------------------------------------------------------------------*/
        if (vm.count("nTimesteps"))
        {
            this->mNtimesteps = vm["nTimesteps"].as<int>();

            if (0 > this->nTimesteps())
            {
                std::cerr << "\nERROR: Number of time steps is negative."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --nTimesteps=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        this->mCoordNodeLast.resize(this->dim());
        this->checkParameter(vm, "coordNodeLast", this->mCoordNodeLast, true);

        /*----------------------------------------------------------------------
        | OPTIONAL PARAMETERS.
        ----------------------------------------------------------------------*/
        this->mCoordNodeFirst.resize(this->dim());
        this->checkParameter(vm, "coordNodeFirst", this->mCoordNodeFirst,
                             false);

        /*--------------------------------------------------------------------*/
        if (vm.count("typeDomainDecomposition"))
        {
            this->mTypeDomainDecomp =
                vm["typeDomainDecomposition"].as<std::string>();
        }
        else
        {
            std::cerr
                << "\nREMARK: Set default --typeDomainDecomposition=RCB.\n\n";
        }

        /*--------------------------------------------------------------------*/
        this->mNpartitions.resize(this->dim());
        if (vm.count("nPartitions"))
        {
            this->checkParameter(vm, "nPartitions", this->mNpartitions, false);
            int nPartitionsTotal = 1;
            for (std::size_t ii = 0; ii < this->nPartitions().size(); ++ii)
            {
                nPartitionsTotal *= (this->mNpartitions[ii]);
            }
            if (nPartitionsTotal != this->mMyWorld.size())
            {
                std::cerr << "\nERROR: #{Partitions} != #{MPI processes}."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nREMARK: Set nPartitions = [nMPIprocs 1,...,1].\n\n";
            /* If the number of partitions is not given at the command line,
            * set the number from the number of generated MPI processes:
            * nPartitions = [MPI_size(), 1, ..., 1]. */
            this->mNpartitions[0] = this->mMyWorld.size();
            for (std::size_t ii = 1; ii < this->nPartitions().size(); ++ii)
            {
                this->mNpartitions[ii] = 1;
            }
        }

        /*--------------------------------------------------------------------*/
        this->mDivideGrid.resize(this->dim());
        this->checkParameter(vm, "divideGrid", this->mDivideGrid, false);

        /*--------------------------------------------------------------------*/
        if (vm.count("readKindFile"))
        {
            this->mKindFile = vm["readKindFile"].as<std::string>();
            this->mReadKindFile = true;
        }
        else
        {
            std::cerr << "\nREMARK: Use --readKindFile=<name>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("writeKindFile"))
        {
            int tmpWriteKindFile = vm["writeKindFile"].as<int>();
            if (0 == tmpWriteKindFile)
            {
                this->mWriteKindFile = false;
            }
            else
            {
                this->mWriteKindFile = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --writeKindFile=<0/1>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("readPartitionFile"))
        {
            this->mNameReadPartitionFile =
                vm["readPartitionfile"].as<std::string>();
            this->mReadPartitionFile = true;
        }
        else
        {
            std::cerr << "\nREMARK: Use --readPartitionfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("writePartitionfile"))
        {
            this->mNameWritePartitionFile =
                vm["writePartitionfile"].as<std::string>();
            this->mWritePartitionFile = true;
        }
        else
        {
            std::cerr << "\nREMARK: Use --writePartitionfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("starttime"))
        {
            this->mTimeIntervalStart = vm["starttime"].as<double>();

            if (0.0 > this->timeIntervalStart())
            {
                std::cerr << "\nERROR: Start time < 0.0."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --starttime=<real value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("endtime"))
        {
            this->mTimeIntervalEnd = vm["endtime"].as<double>();

            if (0.0 >= this->timeIntervalEnd())
            {
                std::cerr << "\nERROR: End time <= 0.0."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --endtime=<real value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("tau"))
        {
            this->mTau = vm["tau"].as<double>();

            if (0.0 >= this->tau())
            {
                std::cerr << "\nERROR: Time step width tau <= 0.0.   "
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else if (vm.count("endtime") && vm.count("nTimesteps"))
        {
            if (0 == this->nTimesteps())
            {
                this->mTau = 0.0;
            }
            else
            {
                this->mTau =
                    (this->timeIntervalEnd() - this->timeIntervalStart()) /
                    this->nTimesteps();
            }
        }
        else
        {
            this->mTau = 1.0;
            std::cout << "REMARK: No time step width tau was specified.\n"
                         " Set tau=1.0 and proceeded. \n\n";
            std::cerr << "\nREMARK: Use --tau=<real value>.\n\n";
        }

        if (vm.count("tau") && vm.count("endTime") && vm.count("nTimesteps"))
        {
            if (this->mTau !=
                (this->timeIntervalEnd() - this->timeIntervalStart()) /
                    this->nTimesteps())
            {
                std::cerr
                    << "\nERROR: Inconsistency between time step width, "
                       "number of time steps and time interval. Exit...\n\n";
                exit(-1);
            }
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("nSnapshots"))
        {
            this->mNsnapshots = vm["nSnapshots"].as<int>();

            if (0 >= nSnapshots())
            {
                std::cerr << "\nERROR: Number of snap shots is less or equal"
                             " to zero. It has to be greater or equal to one."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --nSnapshots=<integral value>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("nLayersAtBorder"))
        {
            this->mNlayersAtBorder = vm["nLayersAtBorder"].as<int>();

            for (std::size_t ii = 0; ii < this->nNodes().size(); ++ii)
            {
                if (this->nLayersAtBorder() >= (this->nNodes()[ii] / 2))
                {
                    std::cerr << "\nERROR: Number of layers at global border"
                                 " is too large (nLayers >= nNodes(ii)/2)."
                                 " Exit...\n\n";
                    this->mMyWorld.abort(-1);
                }
            }
            if (0 >= this->nLayersAtBorder())
            {
                std::cerr << "\nERROR: Number of layers is less or equal"
                             " to zero. It has to be greater or equal to one."
                             " Exit...\n\n";
                this->mMyWorld.abort(-1);
            }
        }
        else
        {
            std::cerr
                << "\nREMARK: Use --nLayersAtBorder=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("outputfile"))
        {
            this->mNameOfDataFile = vm["outputfile"].as<std::string>();
            this->mWriteDataFile = true;
        }
        else
        {
            std::cerr << "\nREMARK: Use --outputfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("configfile"))
        {
            this->mNameOfConfigFile = vm["configfile"].as<std::string>();
        }
        else
        {
            std::cerr << "\nREMARK: Use --configfile=<name>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("enabledAdolc"))
        {
            int tmpEnabledAdolc = vm["enabledAdolc"].as<int>();
            if (0 == tmpEnabledAdolc)
            {
                this->mEnabledAdolc = false;
            }
            else
            {
                this->mEnabledAdolc = true;
#ifndef SCAFES_HAVE_ADOLC
                // Enforce disabling of ADOL-C mode if software was not
                // built with ADOL-C.
                std::cerr << "\nWARNING: Framework was NOT built with ADOL-C "
                             "support!\n";
                std::cerr << "           ===> Enforced enabledAdolc=0!\n\n";

                this->mEnabledAdolc = false;
#endif
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --enabledAdolc=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("asynchronMode"))
        {
            int tmpAsynchronMode = vm["asynchronMode"].as<int>();
            if (0 == tmpAsynchronMode)
            {
                this->mAsynchronMode = false;
            }
            else
            {
                this->mAsynchronMode = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --asynchronMode=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("computeGradients"))
        {
            int tmpComputeGradients = vm["computeGradients"].as<int>();
            if (0 == tmpComputeGradients)
            {
                this->mComputeGradients = false;
            }
            else
            {
                this->mComputeGradients = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --computeGradients=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("checkGradients"))
        {
            int tmpCheckGradients = vm["checkGradients"].as<int>();
            if (0 == tmpCheckGradients)
            {
                this->mCheckGradients = false;
            }
            else
            {
                this->mCheckGradients = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --checkGradients=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("rankOutput"))
        {
            this->mRankOutput = vm["rankOutput"].as<int>();
            if (static_cast<int>(this->rankOutput()) >
                this->myWorld().size() - 1)
            {
                throw std::runtime_error("rankOutput > rankMax.");
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --rankOutput=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("outputlevel"))
        {
            gOutputLevelMax = vm["outputlevel"].as<int>();
            gOutputLevelCurr = gOutputLevelMax;
        }
        else
        {
            std::cerr << "\nREMARK: Use --outputlevel=<integral value>.\n\n";
        }
        } // End of if block related to "help".
    }

    // Communicate result.
    ScaFES::broadcast(this->mMyWorld, *this, 0);
}

/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(Parameters const& rhs)
: mDim(rhs.dim())
, mNnodes(rhs.nNodes())
, mCoordNodeFirst(rhs.coordNodeFirst())
, mCoordNodeLast(rhs.coordNodeLast())
, mNpartitions(rhs.nPartitions())
, mDivideGrid(rhs.divideGrid())
, mReadKindFile(rhs.readKindFile())
, mWriteKindFile(rhs.writeKindFile())
, mKindFile(rhs.kindFile())
, mTypeDomainDecomp(rhs.typeDomainDecomp())
, mWritePartitionFile(rhs.writePartitionFile())
, mNameWritePartitionFile(rhs.nameWritePartitionFile())
, mReadPartitionFile(rhs.readPartitionFile())
, mNameReadPartitionFile(rhs.nameReadPartitionFile())
, mTimeIntervalStart(rhs.timeIntervalStart())
, mTimeIntervalEnd(rhs.timeIntervalEnd())
, mTau(rhs.tau())
, mNtimesteps(rhs.nTimesteps())
, mNsnapshots(rhs.nSnapshots())
, mNlayersAtBorder(rhs.nLayersAtBorder())
, mWriteDataFile(rhs.writeDataFile())
, mNameOfDataFile(rhs.nameDataFile())
, mNameOfConfigFile(rhs.nameConfigFile())
, mEnabledAdolc(rhs.enabledAdolc())
, mAsynchronMode(rhs.asynchronMode())
, mComputeGradients(rhs.computeGradients())
, mCheckGradients(rhs.checkGradients())
, mRankOutput(rhs.rankOutput())
, mMyWorld(rhs.myWorld())
{
}
/*----------------------------------------------------------------------------*/
inline Parameters& Parameters::operator=(Parameters rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
inline Parameters::~Parameters()
{
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
inline int const& Parameters::dim() const
{
    return this->mDim;
}
/*----------------------------------------------------------------------------*/
inline std::vector<int> const& Parameters::nNodes() const
{
    return this->mNnodes;
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeFirst() const
{
    return this->mCoordNodeFirst;
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeLast() const
{
    return this->mCoordNodeLast;
}
/*----------------------------------------------------------------------------*/
inline std::vector<int> const& Parameters::nPartitions() const
{
    return this->mNpartitions;
}
/*----------------------------------------------------------------------------*/
inline std::vector<bool> const& Parameters::divideGrid() const
{
    return this->mDivideGrid;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readKindFile() const
{
    return this->mReadKindFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writeKindFile() const
{
    return this->mWriteKindFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::kindFile() const
{
    return this->mKindFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::typeDomainDecomp() const
{
    return this->mTypeDomainDecomp;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writePartitionFile() const
{
    return this->mWritePartitionFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameWritePartitionFile() const
{
    return this->mNameWritePartitionFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readPartitionFile() const
{
    return this->mReadPartitionFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameReadPartitionFile() const
{
    return this->mNameReadPartitionFile;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalStart() const
{
    return this->mTimeIntervalStart;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalEnd() const
{
    return this->mTimeIntervalEnd;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::tau() const
{
    return this->mTau;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nTimesteps() const
{
    return this->mNtimesteps;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nSnapshots() const
{
    return this->mNsnapshots;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nLayersAtBorder() const
{
    return this->mNlayersAtBorder;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writeDataFile() const
{
    return this->mWriteDataFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameDataFile() const
{
    return this->mNameOfDataFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameConfigFile() const
{
    return this->mNameOfConfigFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::enabledAdolc() const
{
    return this->mEnabledAdolc;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::asynchronMode() const
{
    return this->mAsynchronMode;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::computeGradients() const
{
    return this->mComputeGradients;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::checkGradients() const
{
    return this->mCheckGradients;
}
/*----------------------------------------------------------------------------*/
inline int Parameters::rankOutput() const
{
    return this->mRankOutput;
}
/*----------------------------------------------------------------------------*/
inline ScaFES::Communicator const& Parameters::myWorld() const
{
    return this->mMyWorld;
}
/*----------------------------------------------------------------------------*/
inline int Parameters::rank() const
{
    return this->myWorld().rank();
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::indentDepth() const
{
    return gOutputLevelCurr;
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
inline void Parameters::increaseLevel()
{
    ++gOutputLevelCurr;
}
/*----------------------------------------------------------------------------*/
inline void Parameters::decreaseLevel()
{
    --gOutputLevelCurr;
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
inline bool Parameters::operator==(Parameters const& rhs) const
{
    bool isEqual = true;

    if (this->dim() != rhs.dim())
    {
        isEqual = false;
    }
    if (this->nNodes() != rhs.nNodes())
    {
        isEqual = false;
    }
    if (this->coordNodeFirst() != rhs.coordNodeFirst())
    {
        isEqual = false;
    }
    if (this->coordNodeLast() != rhs.coordNodeLast())
    {
        isEqual = false;
    }
    if (this->nPartitions() != rhs.nPartitions())
    {
        isEqual = false;
    }
    if (this->divideGrid() != rhs.divideGrid())
    {
        isEqual = false;
    }
    if (this->readKindFile() != rhs.readKindFile())
    {
        isEqual = false;
    }
    if (this->kindFile() != rhs.kindFile())
    {
        isEqual = false;
    }
    if (this->writePartitionFile() != rhs.writePartitionFile())
    {
        isEqual = false;
    }
    if (this->nameWritePartitionFile() != rhs.nameWritePartitionFile())
    {
        isEqual = false;
    }
    if (this->nameReadPartitionFile() != rhs.nameReadPartitionFile())
    {
        isEqual = false;
    }
    if (this->timeIntervalStart() != rhs.timeIntervalStart())
    {
        isEqual = false;
    }
    if (this->timeIntervalEnd() != rhs.timeIntervalEnd())
    {
        isEqual = false;
    }
    if (this->tau() != rhs.tau())
    {
        isEqual = false;
    }
    if (this->nTimesteps() != rhs.nTimesteps())
    {
        isEqual = false;
    }
    if (this->nSnapshots() != rhs.nSnapshots())
    {
        isEqual = false;
    }
    if (this->writeDataFile() != rhs.writeDataFile())
    {
        isEqual = false;
    }
    if (this->nameDataFile() != rhs.nameDataFile())
    {
        isEqual = false;
    }
    if (this->nameConfigFile() != rhs.nameConfigFile())
    {
        isEqual = false;
    }
    if (this->enabledAdolc() != rhs.enabledAdolc())
    {
        isEqual = false;
    }
    if (this->computeGradients() != rhs.computeGradients())
    {
        isEqual = false;
    }
    if (this->checkGradients() != rhs.checkGradients())
    {
        isEqual = false;
    }
    if (this->rankOutput() != rhs.rankOutput())
    {
        isEqual = false;
    }
//     if (this->myWorld() != rhs.myWorld())
//     {
//         isEqual = false;
//     }

    return isEqual;
}
/*----------------------------------------------------------------------------*/
inline bool Parameters::operator!=(Parameters const& rhs) const
{
    return !(*this == rhs);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template <typename TT>
inline void Parameters::checkParameter(po::variables_map const& vm,
                                       std::string const& nameParam,
                                       std::vector<TT>& param, bool isNecessary)
{
    if (vm.count(nameParam))
    {
        std::ostringstream ostrRegex;

        for (int ii = 0; ii < dim(); ++ii)
        {
            ostrRegex << "(-?\\d+)";

            if ((dim() - 1) > ii)
            {
                ostrRegex << "x";
            }
        }

        std::string strRegex = ostrRegex.str();
        boost::regex re(strRegex);
        std::string strParam;
        boost::cmatch matches;

        strParam = vm[nameParam].as<std::string>();
        if (boost::regex_match(strParam.c_str(), matches, re))
        {
            for (std::size_t i = 1; i < matches.size(); ++i)
            {
                std::string match(matches[i].first, matches[i].second);
                param.at(i - 1) = atoi(match.c_str());
            }
        }
        else
        {
            std::ostringstream ostrErr;

            for (int ii = 0; ii < dim(); ++ii)
            {
                ostrErr << "n" << ii;

                if ((dim() - 1) > ii)
                {
                    ostrErr << "x";
                }
            }

            std::string strErr = ostrErr.str();
            std::cerr << "\nERROR: The parameter --" << nameParam << "="
                      << strParam << " is not of a format like"
                                     " --" << nameParam << "=" << strErr
                      << ". Exit... \n\n";
            this->mMyWorld.abort(-1);
        }
    }
    else
    {
        std::cerr << "\nREMARK: Use --" << nameParam << "=<a1xa2xa3>\n\n";

        if (isNecessary)
        {
            this->mMyWorld.abort(-1);
        }
    }
}
/*----------------------------------------------------------------------------*/
template <class Archive>
void Parameters::serialize(Archive& ar, const unsigned int version)
{
    if (1 <= version)
    {
        ar&(this->mDim);
        ar&(this->mNnodes);
        ar&(this->mCoordNodeFirst);
        ar&(this->mCoordNodeLast);
        ar&(this->mNpartitions);
        ar&(this->mDivideGrid);
        ar&(this->mReadKindFile);
        ar&(this->mWriteKindFile);
        ar&(this->mKindFile);
        ar&(this->mTypeDomainDecomp);
        ar&(this->mWritePartitionFile);
        ar&(this->mNameWritePartitionFile);
        ar&(this->mReadPartitionFile);
        ar&(this->mNameReadPartitionFile);
        ar&(this->mTimeIntervalStart);
        ar&(this->mTimeIntervalEnd);
        ar&(this->mTau);
        ar&(this->mNtimesteps);
        ar&(this->mNsnapshots);
        ar&(this->mNlayersAtBorder);
        ar&(this->mWriteDataFile);
        ar&(this->mNameOfDataFile);
        ar&(this->mNameOfConfigFile);
        ar&(this->mEnabledAdolc);
        ar&(this->mAsynchronMode);
        ar&(this->mComputeGradients);
        ar&(this->mCheckGradients);
        ar&(this->mRankOutput);
    }
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
inline void swap(Parameters& first, Parameters& second)
{
    std::swap(first.mDim, second.mDim);
    std::swap(first.mNnodes, second.mNnodes);
    std::swap(first.mCoordNodeFirst, second.mCoordNodeFirst);
    std::swap(first.mCoordNodeLast, second.mCoordNodeLast);
    std::swap(first.mNpartitions, second.mNpartitions);
    std::swap(first.mDivideGrid, second.mDivideGrid);
    std::swap(first.mReadKindFile, second.mReadKindFile);
    std::swap(first.mWriteKindFile, second.mWriteKindFile);
    std::swap(first.mKindFile, second.mKindFile);
    std::swap(first.mTypeDomainDecomp, second.mTypeDomainDecomp);
    std::swap(first.mWritePartitionFile, second.mWritePartitionFile);
    std::swap(first.mNameWritePartitionFile, second.mNameWritePartitionFile);
    std::swap(first.mReadPartitionFile, second.mReadPartitionFile);
    std::swap(first.mNameReadPartitionFile, second.mNameReadPartitionFile);
    std::swap(first.mTimeIntervalStart, second.mTimeIntervalStart);
    std::swap(first.mTimeIntervalEnd, second.mTimeIntervalEnd);
    std::swap(first.mTau, second.mTau);
    std::swap(first.mNtimesteps, second.mNtimesteps);
    std::swap(first.mNsnapshots, second.mNsnapshots);
    std::swap(first.mNlayersAtBorder, second.mNlayersAtBorder);
    std::swap(first.mWriteDataFile, second.mWriteDataFile);
    std::swap(first.mNameOfDataFile, second.mNameOfDataFile);
    std::swap(first.mNameOfConfigFile, second.mNameOfConfigFile);
    std::swap(first.mEnabledAdolc, second.mEnabledAdolc);
    std::swap(first.mAsynchronMode, second.mAsynchronMode);
    std::swap(first.mComputeGradients, second.mComputeGradients);
    std::swap(first.mCheckGradients, second.mCheckGradients);
    std::swap(first.mRankOutput, second.mRankOutput);
    std::swap(first.mMyWorld, second.mMyWorld);
}
/*----------------------------------------------------------------------------*/
inline std::ostream& operator<<(std::ostream& output, Parameters const& pm)
{
    if (0 == pm.myWorld().rank())
    {
        output << "* Print parameters... " << std::endl;
        output << "   o Coordinate(left lower node)     = (";

        for (int ii = 0; ii < pm.dim(); ++ii)
        {
            output << boost::format("%4.2f") % pm.coordNodeFirst().at(ii);

            if ((pm.dim() - 1) > ii)
            {
                output << ";";
            }
        }

        output << ")" << std::endl;
        output << "   o Coordinate(right upper node)    = (";

        for (int ii = 0; ii < pm.dim(); ++ii)
        {
            output << boost::format("%4.2f") % pm.coordNodeLast().at(ii);

            if ((pm.dim() - 1) > ii)
            {
                output << ";";
            }
        }

        output << ")" << std::endl;
        output << "   o End time                        = " << std::setw(8)
               << std::scientific << pm.timeIntervalEnd() << std::endl;
        output << "   o #{Time steps}                   = " << std::setw(8)
               << pm.nTimesteps() << std::right << std::endl;
        output << "   o #{Snapshots}                    = " << std::setw(8)
               << pm.nSnapshots() << std::right << std::endl;
        output << "   o Write output file?              = ";

        if (pm.writeDataFile())
        {
            output << std::setw(8) << "YES" << std::right << std::endl;
        }
        else
        {
            output << std::setw(8) << "NO" << std::right << std::endl;
        }

        output << "   o Name of output file =  " << pm.nameDataFile()
               << std::endl;
        output << "  Printed." << std::endl;
    }

    return output;
}

/*******************************************************************************
 * FREE FUNCTIONS.
 ******************************************************************************/
/** Free function which sets a prefix to the output of process with
 * identifier \c rank. */
inline std::string getPrefix(int const& indentationDepth, int const& rank,
                             std::string const& nameFramework =
                                 std::string("ScaFES "))
{
    std::string blanks("     ");
    std::ostringstream blankStream;
    blankStream << nameFramework << ": Rank=" << rank << ": ";
    int nIndents = gOutputLevelMax - indentationDepth;
    for (int jj = 0; jj < nIndents; ++jj)
    {
        blankStream << blanks;
    }
    return blankStream.str();
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
/** Boost serialization version of class \c Parameter. */
BOOST_CLASS_VERSION(ScaFES::Parameters, 2);

#endif
